import { Ref, InjectionKey } from 'vue';

type TurboListener<T> = (payload: T) => void;
interface TurboEvents<E = string> {
    /**
     * Subscribes a given listener.
     */
    subscribe<T = any>(key: E, listener: TurboListener<T>): void;
    /**
     * Unsubscribes the given listener.
     */
    unsubscribe<T = any>(key: E, listener: TurboListener<T>): void;
    /**
     * Emits an event to all active listeners.
     */
    emit<T = any>(key: E, payload: T): void;
}

/**
 * Represents the methods a cache should implement
 * in order to be usable by vue-swr.
 */
interface TurboCache<T = any> {
    /**
     * Gets an item from the cache.
     */
    get<D extends T = T>(key: string): D;
    /**
     * Sets an item to the cache.
     */
    set<D extends T = T>(key: string, value: D): void;
    /**
     * Determines if the cache has a given key.
     */
    has(key: string): boolean;
    /**
     * Removes a key-value pair from the cache.
     */
    remove(key: string): void;
    /**
     * Removes all the key-value pairs from the cache.
     */
    clear(): void;
    /**
     * Returns the current cached keys.
     */
    keys(): string[];
}

/**
 * Determines how we store items in the items cache.
 */
interface ItemsCacheItem<T = any> {
    /**
     * Stores the cache item.
     */
    item: T;
    /**
     * Determines the expiration date of the item.
     */
    expiresAt: Date;
}
/**
 * Determines how we store items in the resolvers cache.
 */
interface ResolversCacheItem<T> {
    /**
     * The resolvable item.
     */
    item: Promise<T>;
    /**
     * The abort controller for the request.
     */
    controller: AbortController;
}
/**
 * Represents the available configuration options
 * for a turbo query instance.
 */
interface TurboQueryConfiguration extends TurboQueryOptions {
    /**
     * Determines the resolved items cache to use.
     */
    itemsCache?: TurboCache<ItemsCacheItem>;
    /**
     * Determines the resolvers cache to use.
     */
    resolversCache?: TurboCache<ResolversCacheItem<any>>;
    /**
     * Stores the event system.
     */
    events?: TurboEvents;
}
interface TurboFetcherAdditional {
    /**
     * An abort signal to cancel pending queries.
     */
    signal: AbortSignal;
}
/**
 * The available options for turbo query.
 */
interface TurboQueryOptions<T = any> {
    /**
     * Determines the item deduplication interval.
     * This determines how many milliseconds an item
     * is considered valid.
     */
    expiration?(item: T): number;
    /**
     * Determines the fetcher function to use.
     */
    fetcher?(key: string, additional: TurboFetcherAdditional): Promise<T>;
    /**
     * Determines if we can return a stale item.
     * If `true`, it will return the previous stale item
     * stored in the cache if it has expired. It will attempt
     * to revalidate it in the background. If `false`, the returned
     * promise will be the revalidation promise.
     */
    stale?: boolean;
    /**
     * Removes the stored item if there is an error in the request.
     * By default, we don't remove the item upon failure, only the resolver
     * is removed from the cache.
     */
    removeOnError?: boolean;
    /**
     * Determines if the result should be a fresh fetched
     * instance regardless of any cached value or its expiration time.
     */
    fresh?: boolean;
}
/**
 * Determines the cache type.
 */
type TurboCacheType = 'resolvers' | 'items';
/**
 * The turbo mutation function type.
 */
type TurboMutateFunction<T> = (old?: T, expiresAt?: Date) => T;
/**
 * The available mutation values.
 */
type TurboMutateValue<T> = T | TurboMutateFunction<T>;
/**
 * Represents the methods a turbo query
 * should implement.
 */
interface TurboQuery {
    /**
     * Configures the current instance of turbo query.
     */
    configure(options?: Partial<TurboQueryConfiguration>): void;
    /**
     * Fetches the key information using a fetcher.
     * The returned promise contains the result item.
     */
    query<T = any>(key: string, options?: TurboQueryOptions<T>): Promise<T>;
    /**
     * Subscribes to a given event on a key. The event handler
     * does have a payload parameter that will contain relevant
     * information depending on the event type.
     */
    subscribe<T = any>(key: string, event: 'refetching', listener: TurboListener<Promise<T>>): () => void;
    subscribe<T = any>(key: string, event: 'resolved', listener: TurboListener<T>): () => void;
    subscribe<T = any>(key: string, event: 'mutated', listener: TurboListener<T>): () => void;
    subscribe<T = any>(key: string, event: 'aborted', listener: TurboListener<Promise<T>>): () => void;
    subscribe<T = any>(key: string, event: 'forgotten', listener: TurboListener<T>): () => void;
    subscribe<T = any>(key: string, event: 'hydrated', listener: TurboListener<T>): () => void;
    subscribe<T = any>(key: string, event: 'error', listener: TurboListener<T>): () => void;
    subscribe<T = any>(key: string, event: TurboQueryEvent, listener: TurboQueryListener<T>): () => void;
    /**
     * Mutates the key with a given optimistic value.
     * The mutated value is considered expired and will be
     * replaced immediatly if a refetch happens.
     */
    mutate<T = any>(key: string, item: TurboMutateValue<T>): void;
    /**
     * Aborts the active resolvers on each key
     * by calling `.abort()` on the `AbortController`.
     * The fetcher is responsible for using the
     * `AbortSignal` to cancel the job.
     */
    abort(key?: string | string[], reason?: any): void;
    /**
     * Forgets the given keys from the cache.
     * Removes items from both, the cache and resolvers.
     */
    forget(keys?: string | string[]): void;
    /**
     * Hydrates the given keys on the cache
     * with the given value and expiration time.
     */
    hydrate<T = any>(keys: string | string[], item: T, expiresAt?: Date): void;
    /**
     * Returns the given keys for the given cache.
     */
    keys(cache?: TurboCacheType): string[];
    /**
     * Returns the expiration date of a given key item.
     * If the item is not in the cache, it will return `undefined`.
     */
    expiration(key: string): Date | undefined;
}
/**
 * Available events on turbo query.
 */
type TurboQueryEvent = 'refetching' | 'resolved' | 'mutated' | 'aborted' | 'forgotten' | 'hydrated' | 'error';
/**
 * Event listeners of turbo query.
 */
type TurboQueryListener<T> = TurboListener<Promise<T> | T>;

declare const configure: (options?: Partial<TurboQueryConfiguration> | undefined) => void;

interface TurboVueOptions extends TurboQueryOptions {
    /**
     * A default turbo query instance to use if any.
     */
    readonly turbo?: TurboQuery;
    /**
     * Determines if it should refetch keys when
     * the window regains focus. You can also
     * set the desired `focusInterval`.
     * @default true
     */
    readonly refetchOnFocus?: boolean;
    /**
     * Determines if it should refetch keys when
     * the connection is restored.
     * @default true
     */
    readonly refetchOnConnect?: boolean;
    /**
     * Determines a throttle interval for the
     * `refetchOnFocus`.
     * @default 5000
     */
    readonly focusInterval?: number;
    /**
     * Clears the resource signal by setting it to
     * undefined when the key is forgotten from the cache.
     * @default false
     */
    readonly clearOnForget?: boolean;
    readonly immediate?: boolean;
}
interface TurboVueResourceActions<T> {
    /**
     * Refetches the current key.
     */
    readonly refetch: () => Promise<T | undefined>;
    /**
     * Mutates the current key.
     */
    readonly mutate: (value: TurboMutateValue<T>) => void;
    /**
     * Usubscribes from the current key changes.
     */
    readonly unsubscribe: () => void;
    /**
     * Forgets the current key from the cache.
     */
    readonly forget: () => void;
    /**
     * Aborts the current key's request if any.
     */
    readonly abort: (reason?: any) => void;
    /**
     * Determines if an error occured.
     */
    readonly error: Readonly<Ref<unknown>>;
    /**
     * Determines if it's refetching in the background.
     */
    readonly isRefetching: Readonly<Ref<boolean>>;
    /**
     * Determines the date of the last window focus.
     * Useful to calculate how many time is left
     * for the next available focus refetching.
     */
    readonly lastFocus: Readonly<Ref<Date>>;
    /**
     * Creates a ref that every given precision interval
     * will determine if the current key is available
     * to refetch via focus and how many time needs to pass till
     * it's available to refetch by focus. This function helps creating
     * the controlled ref on demand rather than creating
     * arbitrary refs ourselves just in case.
     * Return value is [`isAvailable` (readonly ref), `availableIn` (readonly ref)]
     */
    readonly createFocusAvailable: (precision: number) => [Readonly<Ref<boolean>>, Readonly<Ref<number>>];
    /**
     * Determines when the current key expires if
     * it's currently in the cache.
     */
    readonly expiration: () => Date | undefined;
    /**
     * Creates a signal that every given precision interval
     * will determine if the current key is currently expired/stale
     * and how many time needs to pass till its considered expired/stale.
     *
     * This function helps creating the controlled signal on demand
     * rather than creating arbitrary signals ourselves just in case.
     * Return value is `[isStale, staleIn]`
     */
    readonly createStale: (precision: number) => [Readonly<Ref<boolean>>, Readonly<Ref<number>>];
}
type TurboVueResource<T> = [
    /**
     * The resulting resource.
     */
    Readonly<Ref<T | undefined>>,
    /**
     * Available actions on that resource.
     */
    TurboVueResourceActions<T>
];

declare const injectionKey: InjectionKey<TurboVueOptions>;
/**
 * Creates a new query resource with the given key and options
 * for suspense usage.
 */
declare function useAsyncQuery<T = any>(key: () => string | false | null, options?: Omit<TurboVueOptions, 'immediate'>): Promise<TurboVueResource<T>>;
/**
 * Creates a new query resource with the given key and options.
 */
declare function useQuery<T = any>(key: () => string | false | null, options?: TurboVueOptions): TurboVueResource<T>;

export { TurboVueOptions, TurboVueResource, TurboVueResourceActions, configure, injectionKey, useAsyncQuery, useQuery };
