import { inject, computed, ref, watch, getCurrentInstance, onUnmounted, readonly } from 'vue';

function createTurboCache() {
  const items = /* @__PURE__ */ new Map();
  function get(key) {
    if (!items.has(key))
      throw new Error(`The key ${key} is not in the items`);
    return items.get(key);
  }
  function set(key, value) {
    items.set(key, value);
  }
  function has(key) {
    return items.has(key);
  }
  function remove(key) {
    items.delete(key);
  }
  function clear() {
    items.clear();
  }
  function keys() {
    return Array.from(items.keys());
  }
  return { get, set, has, remove, clear, keys };
}

function createTurboEvents() {
  const listeners = /* @__PURE__ */ new Map();
  function subscribe(key, listener) {
    if (!listeners.has(key))
      listeners.set(key, []);
    if (listeners.get(key).includes(listener))
      return;
    listeners.get(key).push(listener);
  }
  function unsubscribe(key, listener) {
    if (!listeners.has(key))
      return;
    if (!listeners.get(key).includes(listener))
      return;
    listeners.get(key).splice(listeners.get(key).indexOf(listener), 1);
    if (listeners.get(key).length === 0)
      listeners.delete(key);
  }
  function emit(key, payload) {
    if (!listeners.has(key))
      return;
    listeners.get(key).forEach((listener) => listener(payload));
  }
  return { subscribe, unsubscribe, emit };
}

function createTurboQuery(instanceOptions) {
  function defaultExpiration() {
    return 2e3;
  }
  async function defaultFetcher() {
    throw new Error("No fetcher defined. Please add a fetcher first.");
  }
  let itemsCache = instanceOptions?.itemsCache ?? createTurboCache();
  let resolversCache = instanceOptions?.resolversCache ?? createTurboCache();
  let events = instanceOptions?.events ?? createTurboEvents();
  let instanceExpiration = instanceOptions?.expiration ?? defaultExpiration;
  let instanceFetcher = instanceOptions?.fetcher ?? defaultFetcher;
  let instanceStale = instanceOptions?.stale ?? true;
  let instanceRemoveOnError = instanceOptions?.removeOnError ?? false;
  let instanceFresh = instanceOptions?.fresh ?? false;
  function configure(options) {
    itemsCache = options?.itemsCache ?? itemsCache;
    resolversCache = options?.resolversCache ?? resolversCache;
    events = options?.events ?? events;
    instanceExpiration = options?.expiration ?? instanceExpiration;
    instanceFetcher = options?.fetcher ?? instanceFetcher;
    instanceStale = options?.stale ?? instanceStale;
    instanceRemoveOnError = options?.removeOnError ?? instanceRemoveOnError;
    instanceFresh = options?.fresh ?? instanceFresh;
  }
  function subscribe(key, event, listener) {
    events.subscribe(`${event}:${key}`, listener);
    if (event === "refetching" && resolversCache.has(key))
      listener(resolversCache.get(key).item);
    return function() {
      events.unsubscribe(`${event}:${key}`, listener);
    };
  }
  function mutate(key, item, expiresAt) {
    if (typeof item === "function") {
      const fn = item;
      const hasKey = itemsCache.has(key);
      const cached = hasKey ? itemsCache.get(key) : void 0;
      item = fn(cached?.item, cached?.expiresAt);
    }
    itemsCache.set(key, { item, expiresAt: expiresAt ?? /* @__PURE__ */ new Date() });
    events.emit(`mutated:${key}`, item);
  }
  function keys(type = "items") {
    return type === "items" ? itemsCache.keys() : resolversCache.keys();
  }
  function abort(cacheKeys, reason) {
    const resolverKeys = typeof cacheKeys === "string" ? [cacheKeys] : cacheKeys ?? keys("resolvers");
    for (const key of resolverKeys) {
      if (resolversCache.has(key)) {
        const resolver = resolversCache.get(key);
        resolver.controller.abort(reason);
        resolversCache.remove(key);
        events.emit(`aborted:${key}`, resolver.item);
      }
    }
  }
  function forget(cacheKeys) {
    const itemKeys = typeof cacheKeys === "string" ? [cacheKeys] : cacheKeys ?? keys("items");
    for (const key of itemKeys) {
      if (itemsCache.has(key)) {
        const item = itemsCache.get(key);
        itemsCache.remove(key);
        events.emit(`forgotten:${key}`, item.item);
      }
    }
  }
  function hydrate(keys2, item, expiresAt) {
    for (const key of typeof keys2 === "string" ? [keys2] : keys2) {
      itemsCache.set(key, { item, expiresAt: expiresAt ?? /* @__PURE__ */ new Date() });
      events.emit(`hydrated:${key}`, item);
    }
  }
  function expiration(key) {
    return itemsCache.has(key) ? itemsCache.get(key).expiresAt : void 0;
  }
  async function query(key, options) {
    const expiration2 = options?.expiration ?? instanceExpiration;
    const fetcher = options?.fetcher ?? instanceFetcher;
    const stale = options?.stale ?? instanceStale;
    const removeOnError = options?.removeOnError ?? instanceRemoveOnError;
    const fresh = options?.fresh ?? instanceOptions?.fresh;
    async function refetch(key2) {
      try {
        if (resolversCache.has(key2))
          return await resolversCache.get(key2).item;
        const controller = new AbortController();
        const result = fetcher(key2, { signal: controller.signal });
        resolversCache.set(key2, { item: result, controller });
        events.emit(`refetching:${key2}`, result);
        const item = await result;
        resolversCache.remove(key2);
        const expiresAt = /* @__PURE__ */ new Date();
        expiresAt.setMilliseconds(
          expiresAt.getMilliseconds() + expiration2(item)
        );
        itemsCache.set(key2, { item, expiresAt });
        events.emit(`resolved:${key2}`, item);
        return item;
      } catch (error) {
        resolversCache.remove(key2);
        if (removeOnError)
          itemsCache.remove(key2);
        events.emit(`error:${key2}`, error);
        throw error;
      }
    }
    if (fresh)
      return await refetch(key);
    if (itemsCache.has(key)) {
      const cached = itemsCache.get(key);
      const hasExpired = cached.expiresAt <= /* @__PURE__ */ new Date();
      if (hasExpired && stale) {
        refetch(key).catch(() => {
        });
        return cached.item;
      }
      if (hasExpired)
        return await refetch(key);
      return cached.item;
    }
    return await refetch(key);
  }
  return {
    query,
    subscribe,
    mutate,
    configure,
    abort,
    forget,
    keys,
    expiration,
    hydrate
  };
}

const {
  query,
  subscribe,
  mutate,
  configure,
  abort,
  forget,
  keys,
  expiration,
  hydrate
} = createTurboQuery();

const injectionKey = Symbol("vue-unquery-context");
async function useAsyncQuery(key, options) {
  const contextOptions = inject(injectionKey);
  const turboQuery = options?.turbo?.query ?? contextOptions?.turbo?.query ?? query;
  const turboMutate = options?.turbo?.mutate ?? contextOptions?.turbo?.mutate ?? mutate;
  const turboSubscribe = options?.turbo?.subscribe ?? contextOptions?.turbo?.subscribe ?? subscribe;
  const turboForget = options?.turbo?.forget ?? contextOptions?.turbo?.forget ?? forget;
  const turboAbort = options?.turbo?.abort ?? contextOptions?.turbo?.abort ?? abort;
  const refetchOnFocus = options?.refetchOnFocus ?? contextOptions?.refetchOnFocus ?? true;
  const refetchOnConnect = options?.refetchOnConnect ?? contextOptions?.refetchOnConnect ?? true;
  const focusInterval = options?.focusInterval ?? contextOptions?.focusInterval ?? 5e3;
  const turboExpiration = options?.turbo?.expiration ?? contextOptions?.turbo?.expiration ?? expiration;
  const clearOnForget = options?.clearOnForget ?? contextOptions?.clearOnForget ?? false;
  const computedKey = computed(() => {
    try {
      return key();
    } catch {
      return null;
    }
  });
  const resource = ref();
  const error = ref();
  const isRefetching = ref(false);
  watch(error, (e) => {
    if (e !== void 0)
      throw e;
  });
  if (computedKey.value) {
    resource.value = await turboQuery(computedKey.value, {
      stale: true,
      ...contextOptions,
      ...options
    });
  }
  async function refetch() {
    if (!computedKey.value)
      return;
    return await turboQuery(computedKey.value, {
      stale: false,
      ...contextOptions,
      ...options
    });
  }
  function localMutate(item) {
    if (!computedKey.value)
      return;
    turboMutate(computedKey.value, item);
  }
  function localForget() {
    if (!computedKey.value)
      return;
    turboForget(computedKey.value);
  }
  function localAbort(reason) {
    if (!computedKey.value)
      return;
    turboAbort(computedKey.value, reason);
  }
  const lastFocus = ref(/* @__PURE__ */ new Date());
  function onFocus(listener) {
    if (!refetchOnFocus || typeof window === "undefined")
      return () => {
      };
    const rawHandler = () => {
      const last = lastFocus.value;
      const now = /* @__PURE__ */ new Date();
      if (now.getTime() - last.getTime() > focusInterval) {
        lastFocus.value = /* @__PURE__ */ new Date();
        listener();
      }
    };
    window.addEventListener("focus", rawHandler);
    return () => window.removeEventListener("focus", rawHandler);
  }
  function onConnect(listener) {
    if (!refetchOnConnect || typeof window === "undefined")
      return () => {
      };
    window.addEventListener("online", listener);
    return () => window.removeEventListener("online", listener);
  }
  function createFocusAvailable(precision) {
    const isAvailable = ref((/* @__PURE__ */ new Date()).getTime() - lastFocus.value.getTime() > focusInterval);
    const availableIn = ref(focusInterval);
    const interval = setInterval(() => {
      const last = lastFocus.value;
      const now = /* @__PURE__ */ new Date();
      const availability = focusInterval - (now.getTime() - last.getTime());
      if (availability >= 0)
        availableIn.value = availability;
      else if (availability < 0 && availableIn.value > 0)
        availableIn.value = 0;
      isAvailable.value = now.getTime() - last.getTime() > focusInterval;
    }, precision);
    onUnmounted(() => {
      clearInterval(interval);
    });
    return [isAvailable, availableIn];
  }
  function localExpiration() {
    const key2 = computedKey.value;
    if (!key2)
      return void 0;
    return turboExpiration(key2);
  }
  function createStale(precision) {
    const now = /* @__PURE__ */ new Date();
    const initialKey = computedKey.value;
    let initialIsStale = true;
    let initialStaleIn = 0;
    if (initialKey) {
      const expiresAt = expiration(initialKey);
      if (expiresAt) {
        const expirationIn = expiresAt.getTime() - now.getTime();
        if (expirationIn >= 0)
          initialStaleIn = expirationIn;
        initialIsStale = expiresAt.getTime() < now.getTime();
      }
    }
    const isStale = ref(initialIsStale);
    const staleIn = ref(initialStaleIn);
    const unsubscribe2 = watch(computedKey, (key2, _old, onCleanup) => {
      const interval = setInterval(() => {
        if (!key2)
          return;
        const expiresAt = expiration(key2);
        if (expiresAt) {
          const now2 = /* @__PURE__ */ new Date();
          const expirationIn = expiresAt.getTime() - now2.getTime();
          if (expirationIn >= 0)
            staleIn.value = expirationIn;
          else if (expirationIn < 0 && staleIn.value > 0)
            staleIn.value = 0;
          isStale.value = expiresAt.getTime() < now2.getTime();
        }
      }, precision);
      onCleanup(() => {
        clearInterval(interval);
      });
    });
    onUnmounted(() => {
      unsubscribe2();
    });
    return [isStale, staleIn];
  }
  const unsubscribe = watch(
    computedKey,
    async (key2, _old, onCleanup) => {
      isRefetching.value = false;
      if (!key2)
        return;
      const unsubscribeMutate = turboSubscribe(
        key2,
        "mutated",
        (item) => {
          resource.value = item;
        }
      );
      const unsubscribeRefetching = turboSubscribe(
        key2,
        "refetching",
        () => {
          isRefetching.value = true;
        }
      );
      const unsubscribeResolved = turboSubscribe(
        key2,
        "resolved",
        (item) => {
          isRefetching.value = false;
          resource.value = item;
        }
      );
      const unsubscribeErrors = turboSubscribe(
        key2,
        "error",
        (e) => {
          isRefetching.value = false;
          error.value = e;
        }
      );
      const unsubscribeForgotten = turboSubscribe(key2, "forgotten", () => {
        if (clearOnForget)
          resource.value = void 0;
      });
      const unsubscribeFocusRefetch = onFocus(() => {
        refetch();
      });
      const unsubscribeConnectRefetch = onConnect(() => {
        refetch();
      });
      resource.value = await turboQuery(key2, {
        stale: true,
        ...contextOptions,
        ...options
      });
      onCleanup(() => {
        unsubscribeMutate();
        unsubscribeRefetching();
        unsubscribeResolved();
        unsubscribeErrors();
        unsubscribeForgotten();
        unsubscribeFocusRefetch();
        unsubscribeConnectRefetch();
      });
    },
    { immediate: true }
  );
  if (getCurrentInstance())
    onUnmounted(() => unsubscribe());
  return [
    // readonly(resource),
    resource,
    {
      refetch,
      mutate: localMutate,
      forget: localForget,
      abort: localAbort,
      unsubscribe,
      error: readonly(error),
      isRefetching: readonly(isRefetching),
      lastFocus: readonly(lastFocus),
      createFocusAvailable,
      expiration: localExpiration,
      createStale
    }
  ];
}
function useQuery(key, options) {
  const contextOptions = inject(injectionKey);
  const turboQuery = options?.turbo?.query ?? contextOptions?.turbo?.query ?? query;
  const turboMutate = options?.turbo?.mutate ?? contextOptions?.turbo?.mutate ?? mutate;
  const turboSubscribe = options?.turbo?.subscribe ?? contextOptions?.turbo?.subscribe ?? subscribe;
  const turboForget = options?.turbo?.forget ?? contextOptions?.turbo?.forget ?? forget;
  const turboAbort = options?.turbo?.abort ?? contextOptions?.turbo?.abort ?? abort;
  const refetchOnFocus = options?.refetchOnFocus ?? contextOptions?.refetchOnFocus ?? true;
  const refetchOnConnect = options?.refetchOnConnect ?? contextOptions?.refetchOnConnect ?? true;
  const focusInterval = options?.focusInterval ?? contextOptions?.focusInterval ?? 5e3;
  const turboExpiration = options?.turbo?.expiration ?? contextOptions?.turbo?.expiration ?? expiration;
  const clearOnForget = options?.clearOnForget ?? contextOptions?.clearOnForget ?? false;
  const immediate = options?.immediate ?? contextOptions?.immediate ?? true;
  const computedKey = computed(() => {
    try {
      return key();
    } catch {
      return null;
    }
  });
  const resource = ref();
  const error = ref();
  const isRefetching = ref(false);
  (async () => {
    if (!computedKey.value || !immediate)
      return;
    resource.value = await turboQuery(computedKey.value, {
      stale: true,
      ...contextOptions,
      ...options
    });
  })();
  async function refetch() {
    if (!computedKey.value)
      return;
    return await turboQuery(computedKey.value, {
      stale: false,
      ...contextOptions,
      ...options
    });
  }
  function localMutate(item) {
    if (!computedKey.value)
      return;
    turboMutate(computedKey.value, item);
  }
  function localForget() {
    if (!computedKey.value)
      return;
    turboForget(computedKey.value);
  }
  function localAbort(reason) {
    if (!computedKey.value)
      return;
    turboAbort(computedKey.value, reason);
  }
  const lastFocus = ref(/* @__PURE__ */ new Date());
  function onFocus(listener) {
    if (!refetchOnFocus || typeof window === "undefined")
      return () => {
      };
    const rawHandler = () => {
      const last = lastFocus.value;
      const now = /* @__PURE__ */ new Date();
      if (now.getTime() - last.getTime() > focusInterval) {
        lastFocus.value = /* @__PURE__ */ new Date();
        listener();
      }
    };
    window.addEventListener("focus", rawHandler);
    return () => window.removeEventListener("focus", rawHandler);
  }
  function onConnect(listener) {
    if (!refetchOnConnect || typeof window === "undefined")
      return () => {
      };
    window.addEventListener("online", listener);
    return () => window.removeEventListener("online", listener);
  }
  function createFocusAvailable(precision) {
    const isAvailable = ref((/* @__PURE__ */ new Date()).getTime() - lastFocus.value.getTime() > focusInterval);
    const availableIn = ref(focusInterval);
    const interval = setInterval(() => {
      const last = lastFocus.value;
      const now = /* @__PURE__ */ new Date();
      const availability = focusInterval - (now.getTime() - last.getTime());
      if (availability >= 0)
        availableIn.value = availability;
      else if (availability < 0 && availableIn.value > 0)
        availableIn.value = 0;
      isAvailable.value = now.getTime() - last.getTime() > focusInterval;
    }, precision);
    onUnmounted(() => {
      clearInterval(interval);
    });
    return [isAvailable, availableIn];
  }
  function localExpiration() {
    const key2 = computedKey.value;
    if (!key2)
      return void 0;
    return turboExpiration(key2);
  }
  function createStale(precision) {
    const now = /* @__PURE__ */ new Date();
    const initialKey = computedKey.value;
    let initialIsStale = true;
    let initialStaleIn = 0;
    if (initialKey) {
      const expiresAt = expiration(initialKey);
      if (expiresAt) {
        const expirationIn = expiresAt.getTime() - now.getTime();
        if (expirationIn >= 0)
          initialStaleIn = expirationIn;
        initialIsStale = expiresAt.getTime() < now.getTime();
      }
    }
    const isStale = ref(initialIsStale);
    const staleIn = ref(initialStaleIn);
    const unsubscribe2 = watch(computedKey, (key2, _old, onCleanup) => {
      const interval = setInterval(() => {
        if (!key2)
          return;
        const expiresAt = expiration(key2);
        if (expiresAt) {
          const now2 = /* @__PURE__ */ new Date();
          const expirationIn = expiresAt.getTime() - now2.getTime();
          if (expirationIn >= 0)
            staleIn.value = expirationIn;
          else if (expirationIn < 0 && staleIn.value > 0)
            staleIn.value = 0;
          isStale.value = expiresAt.getTime() < now2.getTime();
        }
      }, precision);
      onCleanup(() => {
        clearInterval(interval);
      });
    });
    onUnmounted(() => {
      unsubscribe2();
    });
    return [isStale, staleIn];
  }
  const unsubscribe = watch(
    computedKey,
    async (key2, _old, onCleanup) => {
      isRefetching.value = false;
      if (!key2)
        return;
      const unsubscribeMutate = turboSubscribe(
        key2,
        "mutated",
        (item) => {
          resource.value = item;
        }
      );
      const unsubscribeRefetching = turboSubscribe(
        key2,
        "refetching",
        () => {
          isRefetching.value = true;
        }
      );
      const unsubscribeResolved = turboSubscribe(
        key2,
        "resolved",
        (item) => {
          isRefetching.value = false;
          resource.value = item;
        }
      );
      const unsubscribeErrors = turboSubscribe(
        key2,
        "error",
        (e) => {
          isRefetching.value = false;
          error.value = e;
        }
      );
      const unsubscribeForgotten = turboSubscribe(key2, "forgotten", () => {
        if (clearOnForget)
          resource.value = void 0;
      });
      const unsubscribeFocusRefetch = onFocus(() => {
        refetch();
      });
      const unsubscribeConnectRefetch = onConnect(() => {
        refetch();
      });
      resource.value = await turboQuery(key2, {
        stale: true,
        ...contextOptions,
        ...options
      });
      onCleanup(() => {
        unsubscribeMutate();
        unsubscribeRefetching();
        unsubscribeResolved();
        unsubscribeErrors();
        unsubscribeForgotten();
        unsubscribeFocusRefetch();
        unsubscribeConnectRefetch();
      });
    },
    { immediate }
  );
  if (getCurrentInstance())
    onUnmounted(() => unsubscribe());
  return [
    // readonly(resource),
    resource,
    {
      refetch,
      mutate: localMutate,
      forget: localForget,
      abort: localAbort,
      unsubscribe,
      error: readonly(error),
      isRefetching: readonly(isRefetching),
      lastFocus: readonly(lastFocus),
      createFocusAvailable,
      expiration: localExpiration,
      createStale
    }
  ];
}

export { configure, injectionKey, useAsyncQuery, useQuery };
